//
//  LKFMetaDataGroup.swift
//  LKF
//
//  Created by Alexander Holzer on 17.01.22.
//  Copyright © 2022 Alexander Holzer. All rights reserved.
//

import Foundation
@_implementationOnly import LibLKF

public class LKFMetaDataGroupNode {
    static func nodeForIndex (context: LKFContext, ptr: OpaquePointer, index: UInt32) -> LKFMetaDataGroupNode? {
        if let contextptr = context.ptr {
            let lineBuffer = UnsafeMutablePointer<CChar>.allocate(capacity:1024)
            if LibLKF.lkf_stamm_getGroupLine (contextptr, ptr, index, lineBuffer, 1024) > 0 {
                let r = String(cString: lineBuffer)
                lineBuffer.deallocate()
                return nodeForIndex(context: context, ptr: ptr, definition: r,index: index)
            }
        }
        return nil
    }
    
    static func nodeForIndex (context: LKFContext, ptr: OpaquePointer, definition: String, index: UInt32) -> LKFMetaDataGroupNode? {
        if definition.starts(with: "L") {
            return LKFMetaDataGroupLeafNode (context: context, index: index, definition: definition)
        }
        return LKFMetaDataGroupSplitNode (context: context, index: index, ptr: ptr, definition: definition)
    }
    
    public func parallelChildrenCount (maxDepth: Int = 99) -> Int {
        return 1
    }
    
    public var depth : Int {
        return 0
    }
}

public class LKFMetaDataGroupLeafNode : LKFMetaDataGroupNode {
    public let knoten : String
    public let tk : Double
    public let lk : Double
    public let melzu : Double
    public let vdu: Int
    public let vdo: Int
    public let vdmw : Double

    init? (context: LKFContext, index: UInt32, definition: String) {
        let parts = definition.split(separator: "|")
        if parts.count > 8, let tk = Double (parts[2]), let lk = Double (parts[3]) {
            self.tk = tk
            self.lk = lk
            self.knoten = String(parts[1])
            self.melzu  = Double(parts[4]) ?? 0.0
            self.vdu = Int(parts[6]) ?? 0
            self.vdo = Int(parts[7]) ?? 0
            self.vdmw = Double(parts[8]) ?? 0.0
        }
        else {
            return nil
        }
        
        
    }
}

public class LKFMetaDataGroupSplitNode : LKFMetaDataGroupNode {
    public let trueNode : LKFMetaDataGroupNode
    public let falseNode : LKFMetaDataGroupNode
    public let nodeType : Int
    public let compare : String
    public let intervalOperator : String?
    public let interval0 : Int?
    public let interval1 : Int?
    
    init? (context: LKFContext, index: UInt32, ptr: OpaquePointer, definition: String) {
        let parts = definition.split(separator: "|")
        if parts.count > 4, let falseIndex = UInt32(parts[1]), let trueIndex = UInt32(parts[2]),
           let nodeType = Int(parts[3]),
           let trueNode = LKFMetaDataGroupNode.nodeForIndex(context: context, ptr: ptr, index: trueIndex),
           let falseNode = LKFMetaDataGroupNode.nodeForIndex(context: context, ptr: ptr, index: falseIndex) {
            self.trueNode = trueNode
            self.falseNode = falseNode
            self.compare = String(parts[4])
            self.nodeType = nodeType
        }
        else {
            return nil
        }
        
        if parts.count > 6 {
            self.intervalOperator = String(parts[5])
            self.interval0 = Int(parts[6])
            if parts.count > 7 {
                self.interval1 = Int(parts[7])
            }
            else {
                self.interval1 = nil
            }

        }
        else {
            self.intervalOperator = nil
            self.interval0 = nil
            self.interval1 = nil
        }
    }
    
    public var nodeTypeHasInterval : Bool {
        switch (nodeType) {
        case 1, 2, 3, 4, 5, 7, 12:
            return false
        default:
            return true
        }
    }
    
    public override func parallelChildrenCount (maxDepth: Int = 99) -> Int {
        if maxDepth > 0 {
            return trueNode.parallelChildrenCount(maxDepth: maxDepth - 1) + falseNode.parallelChildrenCount(maxDepth: maxDepth - 1)
        }
        return 1
    }
    
    public override var depth: Int {
        return max(trueNode.depth + 1, falseNode.depth + 1)
    }
    
    public var nodeTypeText : String {
        switch (nodeType) {
        case 1: return "SPEZMEL"
        case 2: return "SPEZHDG4"
        case 3: return "SPEZHDG"
        case 4: return "SPEZDG4"
        case 5: return "SPEZDG"
        case 6: return "ALTER"
        case 7: return "AGR01"
        case 12: return "AGR02"
       default:
            if nodeType >= 101 && nodeType <= 199 {
                return String(format:"HG%03d", nodeType)
            }
            if nodeType > 300 && nodeType <= 400 {
                return String(format:"ALG%02d", nodeType - 300)
            }
            if nodeType > 400 && nodeType <= 500 {
                return String(format:"AFG%02d", nodeType - 400)
            }
            if nodeType <= 699 {
                return String(format:"LG%03d", nodeType - 400)
            }
            return "\(nodeType)"
            
        }
    }
    
    public var intervalString : String? {
        if nodeTypeHasInterval, let op = intervalOperator {
            let interval0String = interval0 != nil ? String(interval0!) : ""
            if op == "><" {
                let interval1String = interval1 != nil ? String(interval1!) : ""
                return "\(interval0String) - \(interval1String)"
            }
            else {
            
                return "\(op)\(interval0String)"
            }
        }
        return nil
    }
}

public class LKFMetaDataGroup {
    public let group: String
    let ptr: OpaquePointer
    public let rootNode : LKFMetaDataGroupNode
    
    public init? (context: LKFContext, group: String) {
        
        guard let ptr = context.ptr else {
            return nil
        }
        
        guard let p = LibLKF.lkf_stamm_getGroup (ptr, group.cString(using: .utf8)) else {
            return nil
        }
        
        guard let rootNode = LKFMetaDataGroupNode.nodeForIndex(context: context, ptr: p, index: 0) else {
            return nil
        }
        self.group = group
        self.ptr = p
        self.rootNode = rootNode
    }
}
